@@ -18,7 +18,7 @@ limitations under the License.
 
 #include <vector>
 
-#include "third_party/eigen3/Eigen/Core"
+#include "Eigen/Core"  // from @eigen_archive
 #include "tensorflow/core/framework/device_base.h"
 #include "tensorflow/core/lib/core/errors.h"
 #include "tensorflow/core/lib/core/status.h"
@@ -279,7 +279,7 @@ Status CTCLossCalculator<T>::CalculateLoss(
   } else {
     ComputeLossAndGradients(0, batch_size);
   }
-  return OkStatus();
+  return absl::OkStatus();
 }
 
 template <class T>
@@ -363,7 +363,7 @@ Status CTCLossCalculator<T>::PopulateLPrimes(
     GetLPrimeIndices(l, &l_primes->at(b));
     *max_u_prime = std::max(*max_u_prime, l_primes->at(b).size());
   }
-  return OkStatus();
+  return absl::OkStatus();
 }
 
 // Calculates the alpha(t, u) as described in (GravesTh) Section 7.3.
